"""
Export colors in various formats.
"""

import logging
import os
import re

from . import util
from .settings import CACHE_DIR, CONF_DIR, MODULE_DIR


class ExportFile:
    """A simple class for representing the few things
    needed to read a file and exporting it.
    """

    def __init__(self, abs_path, base_dir):
        self.name = os.path.basename(abs_path)
        self.path = abs_path
        self.relative_path = os.path.relpath(abs_path, base_dir)


def template(colors, input_file, output_file=None):
    """Read template file, substitute markers and
    save the file elsewhere."""
    # pylint: disable-msg=too-many-locals
    template_data = util.read_file_raw(input_file)
    for i, l in enumerate(template_data):
        for match in re.finditer(r"(?<=(?<!\{))(\{([^{}]+)\})(?=(?!\}))", l):
            # Get the color, and the functions associated with it
            cname, _, funcs = match.group(2).partition(".")
            # Check that functions are needed for this color
            if len(funcs) == 0:
                continue
            # Build up a string which will be replaced with the new color
            replace_str = cname
            # Color to be modified copied into new one
            new_color = util.Color(colors[cname].hex_color)
            # Execute each function to be done
            for func in filter(None, re.split(r"\)|\.", funcs)):
                # Get function name and arguments
                func = func.split("(")
                fname = func[0]
                if fname[0] == ".":
                    fname = fname[1:]
                if not hasattr(new_color, fname):
                    logging.error(
                        "Syntax error in template file '%s' on line '%s'",
                        input_file,
                        i,
                    )
                function = getattr(new_color, fname)

                # If the function is callable, call it
                if callable(function):
                    if len(func) > 1:
                        new_color = function(*func[1].split(","))
                    else:
                        new_color = function()
                    # string to replace generated colors
                    if func[0] != ".":
                        replace_str += "."
                    replace_str += "(".join(func) + ")"
                else:
                    # if it is an attribute i.e. rgb
                    replace_str += "." + fname
                    new_color = function

            if isinstance(new_color, util.Color):
                new_color = new_color.strip
            # If replace the format placeholder with the new color
            if new_color is not colors[cname]:
                new_color = str(new_color)
                template_data[i] = l.replace("{" + replace_str + "}", new_color)
    try:
        template_data = "".join(template_data).format(**colors)
    except (ValueError, KeyError, AttributeError) as exc:
        logging.error(
            "Syntax error in template file '%s': %r.", input_file, exc
        )
        return
    util.save_file(template_data, output_file)


def flatten_colors(colors):
    """Prepare colors to be exported.
    Flatten dicts and convert colors to util.Color()"""
    all_colors = {
        "wallpaper": colors["wallpaper"],
        "checksum": colors["checksum"],
        "alpha": colors["alpha"],
        **colors["special"],
        **colors["colors"],
    }
    return {k: util.Color(v) for k, v in all_colors.items()}


def get_export_type(export_type):
    """Convert template type to the right filename."""
    return {
        "css": "colors.css",
        "dmenu": "colors-wal-dmenu.h",
        "dwm": "colors-wal-dwm.h",
        "dwm_urg": "colors-wal-dwm-urg.h",
        "st": "colors-wal-st.h",
        "tabbed": "colors-wal-tabbed.h",
        "gtk2": "colors-gtk2.rc",
        "json": "colors.json",
        "konsole": "colors-konsole.colorscheme",
        "kitty": "colors-kitty.conf",
        "nqq": "colors-nqq.css",
        "plain": "colors",
        "putty": "colors-putty.reg",
        "rofi": "colors-rofi.Xresources",
        "scss": "colors.scss",
        "shell": "colors.sh",
        "fishshell": "colors.fish",
        "speedcrunch": "colors-speedcrunch.json",
        "sway": "colors-sway",
        "tty": "colors-tty.sh",
        "vscode": "colors-vscode.json",
        "waybar": "colors-waybar.css",
        "polybar": "colors-polybar",
        "xresources": "colors.Xresources",
        "haskell": "Colors.hs",
        "xmonad": "colors.hs",
        "yaml": "colors.yml",
    }.get(export_type, export_type)


def walk(directory):
    """Walks a directory tree and yields files for export"""
    for root, _, files in os.walk(directory):
        for file in files:
            yield (ExportFile(os.path.join(root, file), directory))


def every(colors, output_dir=CACHE_DIR):
    """Export all template files."""
    join = os.path.join  # Minor optimization.
    colors = flatten_colors(colors)
    template_dir = join(MODULE_DIR, "templates")
    template_dir_user = join(CONF_DIR, "templates")
    util.create_dir(template_dir_user)

    logging.info("Reading system templates from: %s", template_dir)
    logging.info("Reading user templates from: %s", template_dir_user)
    for file in [*walk(template_dir), *walk(template_dir_user)]:
        if file.name != ".DS_Store" and not file.name.endswith(".swp"):
            template(colors, file.path, join(output_dir, file.relative_path))

    logging.info("Exported all files.")
    logging.info("Exported all user files to %s", output_dir)


def color(colors, export_type, output_file=None):
    """Export a single template file."""
    all_colors = flatten_colors(colors)

    template_name = get_export_type(export_type)

    if template_name == export_type:
        template_file = os.path.join(CONF_DIR, "templates", template_name)
    else:
        template_file = os.path.join(MODULE_DIR, "templates", template_name)

    output_file = output_file or os.path.join(CACHE_DIR, template_name)

    if os.path.isfile(template_file):
        template(all_colors, template_file, output_file)
        logging.info("Exported %s.", export_type)
    else:
        logging.warning("Template '%s' doesn't exist.", export_type)



"""
Generate a colorscheme using okthief.
"""

import logging
import shutil
import subprocess
import sys
import json

from .. import colors
from .. import util


def gen_colors(img):
    """Generate a colorscheme using okthief."""
    cmd = ["okthief", "--num-colors", "8", "--file"]
    out = subprocess.check_output([*cmd, img])
    cols = [x["hex"] for x in json.loads(out)]
    white, black, rest = (cols[0], cols[1], cols[2:])
    cols = [black]
    cols.extend(rest)
    cols.append(white)
    cols.extend(cols.copy())
    return cols


def adjust(cols, light, **kwargs):
    """Create palette.
    :keyword-args:
    -    c16: use 16 colors through specified method - [ "lighten" | "darken" ]
    """
    if "c16" in kwargs:
        cols16 = kwargs["c16"]
    else:
        cols16 = False
    return colors.generic_adjust(cols, light, c16=cols16)


def get(img, light=False, **kwargs):
    """Get colorscheme.
    :keyword-args:
    -    c16: use 16 colors through specified method - [ "lighten" | "darken" ]
    """
    if "c16" in kwargs:
        cols16 = kwargs["c16"]
    else:
        cols16 = False
    if not shutil.which("okthief"):
        logging.error("okthief wasn't found on your system.")
        logging.error("Try another backend. (wal --backend)")
        sys.exit(1)

    cols = gen_colors(img)
    cols[0] = util.darken_color(cols[0], 0.80)
    return adjust(cols, light, c16=cols16)



"""
Reload programs.
"""

import logging
import os
import shutil
import subprocess
import time

from .settings import CACHE_DIR, MODULE_DIR, OS
from . import util


def tty(tty_reload):
    """Load colors in tty."""
    tty_script = os.path.join(CACHE_DIR, "colors-tty.sh")
    term = os.environ.get("TERM")

    if tty_reload and term == "linux":
        subprocess.Popen(["sh", tty_script])


def xrdb(xrdb_files=None):
    """Merge the colors into the X db so new terminals use them."""
    xrdb_files = xrdb_files or [os.path.join(CACHE_DIR, "colors.Xresources")]

    if shutil.which("xrdb") and OS != "Darwin":
        for file in xrdb_files:
            subprocess.run(["xrdb", "-merge", "-quiet", file], check=False)


def i3():
    """Reload i3 colors."""
    if shutil.which("i3-msg") and util.get_pid("i3"):
        util.disown(["i3-msg", "reload"])


def bspwm():
    """Reload bspwm colors."""
    if shutil.which("bspc") and util.get_pid("bspwm"):
        util.disown(["bspc", "wm", "-r"])


def kitty():
    """Reload kitty colors."""
    if shutil.which("kitty") and util.get_pid("kitty"):
        if os.getenv("TERM") == "xterm-kitty":
            subprocess.call(
                [
                    "kitty",
                    "@",
                    "set-colors",
                    "--all",
                    os.path.join(CACHE_DIR, "colors-kitty.conf"),
                ]
            )
        else:
            subprocess.call(
                [
                    "kitty",
                    "@",
                    "--to",
                    "unix:/tmp/kitty_pywal",
                    "set-colors",
                    "--all",
                    os.path.join(CACHE_DIR, "colors-kitty.conf"),
                ]
            )


def polybar():
    """Reload polybar colors."""
    if shutil.which("polybar") and util.get_pid("polybar"):
        util.disown(["pkill", "-USR1", "polybar"])


def sway():
    """Reload sway colors."""
    if shutil.which("swaymsg") and util.get_pid("sway"):
        util.disown(["swaymsg", "reload"])


def firefox():
    """reload pywalfox."""
    if shutil.which("pywalfox"):
        util.disown(["pywalfox", "update"])


def waybar():
    """Reload waybar colors."""
    if shutil.which("waybar") and util.get_pid("waybar"):
        util.disown(["pkill", "-USR2", "waybar"])


def colors(cache_dir=CACHE_DIR):
    """Reload colors. (Deprecated)"""
    sequences = os.path.join(cache_dir, "sequences")

    logging.error("'wal -r' is deprecated: " "Use 'cat %s' instead.", sequences)

    if os.path.isfile(sequences):
        print("".join(util.read_file(sequences)), end="")


def termux():
    """reload termux colors."""
    if shutil.which("termux-reload-settings"):
        util.disown(["termux-reload-settings"])


def mako():
    """Reload mako colors."""
    if shutil.which("mako") and util.get_pid("mako"):
        util.disown(["makoctl", "reload"])


def nvim():
    """Reload nvim colors."""
    if shutil.which("nvim-colo-reload") and util.get_pid("nvim"):
        util.disown(["nvim-colo-reload"])


def env(xrdb_file=None, tty_reload=True):
    """Reload environment."""
    xrdb(xrdb_file)
    i3()
    bspwm()
    kitty()
    sway()
    polybar()
    nvim()
    waybar()
    termux()
    mako()
    firefox()
    logging.info("Reloaded environment.")
    tty(tty_reload)



"""
Generate a colorscheme using Colorz.
"""

import logging
import sys

try:
    import colorz

except ImportError:
    logging.error("colorz wasn't found on your system.")
    logging.error("Try another backend. (wal --backend)")
    sys.exit(1)

from .. import colors
from .. import util


def gen_colors(img):
    """Generate a colorscheme using Colorz."""
    # pylint: disable=not-callable
    raw_colors = colorz.colorz(img, n=6, bold_add=0)
    return [util.rgb_to_hex([*color[0]]) for color in raw_colors]


def adjust(cols, light, **kwargs):
    """Create palette.
    :keyword-args:
    -    c16: use 16 colors through specified method - [ "lighten" | "darken" ]
    """
    if "c16" in kwargs:
        cols16 = kwargs["c16"]
    else:
        cols16 = False
    raw_colors = [cols[0], *cols, "#FFFFFF", "#000000", *cols, "#FFFFFF"]
    raw_colors[0] = util.darken_color(cols[0], 0.80)

    return colors.generic_adjust(raw_colors, light, c16=cols16)


def get(img, light=False, **kwargs):
    """Get colorscheme.
    :keyword-args:
    -    c16: use 16 colors through specified method - [ "lighten" | "darken" ]
    """
    if "c16" in kwargs:
        cols16 = kwargs["c16"]
    else:
        cols16 = False
    cols = gen_colors(img)

    if len(cols) < 6:
        logging.error("colorz failed to generate enough colors.")
        logging.error("Try another backend or another image. (wal --backend)")
        sys.exit(1)

    return adjust(cols, light, c16=cols16)



"""
Generate a colorscheme using fast_colorthief.
"""

import logging
import sys

try:
    import fast_colorthief

except ImportError:
    logging.error("fast_colorthief wasn't found on your system.")
    logging.error("Try another backend. (wal --backend)")
    sys.exit(1)

from .. import util
from .. import colors


def gen_colors(img):
    """Ask backend to generate 16 colors."""
    raw_colors = fast_colorthief.get_palette(img, 16)

    return [util.rgb_to_hex(color) for color in raw_colors]


def adjust(cols, light, **kwargs):
    """Create palette.
    :keyword-args:
    -    c16: use 16 colors through specified method - [ "lighten" | "darken" ]
    """
    if "c16" in kwargs:
        cols16 = kwargs["c16"]
    else:
        cols16 = False
    cols.sort(key=util.rgb_to_yiq)
    raw_colors = [*cols, *cols]
    raw_colors[0] = util.darken_color(cols[0], 0.80)

    return colors.generic_adjust(raw_colors, light, c16=cols16)


def get(img, light=False, **kwargs):
    """Get colorscheme.
    :keyword-args:
    -    c16: use 16 colors through specified method - [ "lighten" | "darken" ]
    """
    if "c16" in kwargs:
        cols16 = kwargs["c16"]
    else:
        cols16 = False
    cols = gen_colors(img)
    return adjust(cols, light, c16=cols16)



"""Set the wallpaper."""

import ctypes
import logging
import os
import re
import shutil
import subprocess
import urllib.parse
import plistlib
import datetime
import tempfile

from .settings import HOME, OS, CACHE_DIR
from . import util


def get_desktop_env():
    """Identify the current running desktop environment."""
    desktop = os.environ.get("XDG_CURRENT_DESKTOP")
    if desktop:
        return desktop

    desktop = os.environ.get("DESKTOP_SESSION")
    if desktop:
        return desktop

    desktop = os.environ.get("GNOME_DESKTOP_SESSION_ID")
    if desktop:
        return "GNOME"

    desktop = os.environ.get("MATE_DESKTOP_SESSION_ID")
    if desktop:
        return "MATE"

    desktop = os.environ.get("SWAYSOCK")
    if desktop:
        return "SWAY"

    desktop = os.environ.get("DESKTOP_STARTUP_ID")
    if desktop and "awesome" in desktop:
        return "AWESOME"

    return None


def xfconf(img):
    """Call xfconf to set the wallpaper on XFCE."""
    xfconf_re = re.compile(
        r"^/backdrop/screen\d/monitor(?:0|\w*)/"
        r"(?:(?:image-path|last-image)|workspace\d/last-image)$",
        flags=re.M,
    )
    xfconf_data = subprocess.check_output(
        ["xfconf-query", "--channel", "xfce4-desktop", "--list"],
        stderr=subprocess.DEVNULL,
    ).decode("utf8")
    paths = xfconf_re.findall(xfconf_data)
    for path in paths:
        util.disown(
            [
                "xfconf-query",
                "--channel",
                "xfce4-desktop",
                "--property",
                path,
                "--set",
                img,
            ]
        )


def set_wm_wallpaper(img):
    """Set the wallpaper for non desktop environments."""
    if shutil.which("swww"):
        util.disown(["swww", "img", img])

    elif shutil.which("swaybg"):
        subprocess.call(["killall", "swaybg"])
        util.disown(["swaybg", "-m", "fill", "-i", img])

    elif shutil.which("feh"):
        util.disown(["feh", "--bg-fill", img])

    elif shutil.which("wbg"):
        subprocess.call(["killall", "wbg"])
        util.disown(["wbg", img])

    elif shutil.which("xwallpaper"):
        util.disown(["xwallpaper", "--zoom", img])

    elif shutil.which("nitrogen"):
        util.disown(["nitrogen", "--set-zoom-fill", img])

    elif shutil.which("bgs"):
        util.disown(["bgs", "-z", img])

    elif shutil.which("hsetroot"):
        util.disown(["hsetroot", "-fill", img])

    elif shutil.which("habak"):
        util.disown(["habak", "-mS", img])

    elif shutil.which("display"):
        util.disown(["display", "-backdrop", "-window", "root", img])

    else:
        logging.error("No wallpaper setter found.")
        return


def set_desktop_wallpaper(desktop, img):
    """Set the wallpaper for the desktop environment."""
    desktop = str(desktop).lower()

    if "xfce" in desktop or "xubuntu" in desktop:
        xfconf(img)

    elif "muffin" in desktop or "cinnamon" in desktop:
        util.disown(
            [
                "gsettings",
                "set",
                "org.cinnamon.desktop.background",
                "picture-uri",
                "file://" + urllib.parse.quote(img),
            ]
        )

    elif "gnome" in desktop or "unity" in desktop:
        util.disown(
            [
                "gsettings",
                "set",
                "org.gnome.desktop.background",
                "picture-uri-dark",
                "file://" + urllib.parse.quote(img),
            ]
        )
        util.disown(
            [
                "gsettings",
                "set",
                "org.gnome.desktop.background",
                "picture-uri",
                "file://" + urllib.parse.quote(img),
            ]
        )

    elif "mate" in desktop:
        util.disown(
            ["gsettings", "set", "org.mate.background", "picture-filename", img]
        )

    elif "sway" in desktop:
        util.disown(["swaymsg", "output", "*", "bg", img, "fill"])

    elif "awesome" in desktop:
        util.disown(
            [
                "awesome-client",
                "require('gears').wallpaper.maximized('{img}')".format(
                    **locals()
                ),
            ]
        )

    elif "kde" in desktop:
        string = """
            var allDesktops = desktops();for (i=0;i<allDesktops.length;i++){
            d = allDesktops[i];d.wallpaperPlugin = "org.kde.image";
            d.currentConfigGroup = Array("Wallpaper", "org.kde.image",
            "General");d.writeConfig("Image", "%s")};
        """
        if shutil.which("qdbus6"):
            use_qdbus = "qdbus6"
        elif shutil.which("qdbus5"):
            use_qdbus = "qdbus5"
        elif shutil.which("qdbus"):
            use_qdbus = "qdbus"
        else:
            logging.error("No qdbus6, qdbus5 or qdbus binary found")
            return
        util.disown(
            [
                use_qdbus,
                "org.kde.plasmashell",
                "/PlasmaShell",
                "org.kde.PlasmaShell.evaluateScript",
                string % img,
            ]
        )

    elif "hyprland" in desktop and shutil.which("hyprpaper"):
        util.disown(["hyprctl", "hyprpaper", "preload ", img])
        util.disown(["hyprctl", "hyprpaper", "wallpaper", ", " + img])
    else:
        set_wm_wallpaper(img)


def set_mac_wallpaper(img):
    """Set the wallpaper on macOS."""
    db_file = "Library/Application Support/Dock/desktoppicture.db"
    db_path = os.path.join(HOME, db_file)

    # Fresh installs of Sonoma will not have this file.
    # Check if file exists to make backwards compatibility forwards compatible.
    if os.path.isfile(db_path):

        # Put the image path in the database
        sql = 'insert into data values("%s"); ' % img
        subprocess.call(["sqlite3", db_path, sql])

        # Get the index of the new entry
        sql = "select max(rowid) from data;"
        new_entry = subprocess.check_output(["sqlite3", db_path, sql])
        new_entry = new_entry.decode("utf8").strip("\n")

        # Get all picture ids (monitor/space pairs)
        get_pics_cmd = ["sqlite3", db_path, "select rowid from pictures;"]
        pictures = subprocess.check_output(get_pics_cmd)
        pictures = pictures.decode("utf8").split("\n")

        # Clear all existing preferences
        sql += "delete from preferences; "

        # Write all pictures to the new image
        for pic in pictures:
            if pic:
                sql += "insert into preferences (key, data_id, picture_id) "
                sql += "values(1, %s, %s); " % (new_entry, pic)

        subprocess.call(["sqlite3", db_path, sql])

        # Kill the dock to fix issues with cached wallpapers.
        # macOS caches wallpapers and if a wallpaper is set that shares
        # the filename with a cached wallpaper, the cached wallpaper is
        # used instead.
        subprocess.call(["killall", "Dock"])

    # MacOS Sonomoa uses a plist file instead.  Interestingly the database
    # referenced above still exists, but doesn't seem to do anyything on Sonoma
    # so lets leave it for backward compatatility

    plist_path = (
        "Library/Application Support/com.apple.wallpaper/Store/Index.plist"
    )
    plist_path = os.path.join(HOME, plist_path)

    # Write a backup of plist file in temp in case something horrific happens
    with open(plist_path, "rb") as f:
        old_plist = plistlib.load(f)
        with tempfile.NamedTemporaryFile(
            prefix="pywal-plist-bk-", delete=False
        ) as g:
            logging.info(f"Backup plist file saved to {g.name}")
            plistlib.dump(old_plist, g)

    # Unfortunately these fields seem mandatory.  not extensively tested
    # - Configuration is a plist unto itself and a value is required
    with open(plist_path, "wb") as f:
        new_plist = {
            "Spaces": {},
            "SystemDefault": {
                "Desktop": {
                    "LastSet": datetime.datetime.now(),
                    "LastUse": datetime.datetime.now(),
                    "Content": {
                        "Choices": [
                            {
                                "Provider": "com.apple.wallpaper.choice.image",
                                "Files": [{"relative": f"file:///{img}"}],
                                "Configuration": b"",
                            }
                        ],
                        "Shuffle": "$null",
                    },
                },
                "Type": "individual",
                "Idle": {
                    "LastSet": datetime.datetime.now(),
                    "LastUse": datetime.datetime(
                        2023, 10, 21, 2, 51, 4, 435303
                    ),
                    "Content": {
                        "Choices": [
                            {
                                "Provider": "com.apple.wallpaper.choice.aerials",
                                "Files": [],
                                "Configuration": b"",
                            }
                        ],
                        "Shuffle": {
                            "Type": "afterDuration",
                            "Duration": [2341, 16172123445939666944],
                        },
                    },
                },
            },
            "Displays": {},
            "AllSpacesAndDisplays": {
                "Desktop": {
                    "LastSet": datetime.datetime.now(),
                    "LastUse": datetime.datetime.now(),
                    "Content": {
                        "Choices": [
                            {
                                "Provider": "com.apple.wallpaper.choice.image",
                                "Files": [{"relative": f"file:///{img}"}],
                                "Configuration": b'bplist00\xd2\x01\x02\x03\x0c_\x10\x0fbackgroundColorYplacement\xd2\x04\x05\x06\x0bZcomponentsZcolorSpace\xa4\x07\x08\t\n#?\xd0PPPPPP#?\xdaZZZZZZ#?\xe5UUUUUU#?\xf0\x00\x00\x00\x00\x00\x00O\x10Cbplist00_\x10\x17kCGColorSpaceGenericRGB\x08\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"\x10\x01\x08\r\x1f).9DIR[dm\xb3\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\r\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb5',
                            }
                        ],
                        "Shuffle": "$null",
                    },
                    "Last": datetime.datetime.now(),
                },
                "Type": "individual",
                "Idle": {
                    "LastSet": datetime.datetime.now(),
                    "LastUse": datetime.datetime.now(),
                    "Content": {
                        "Choices": [
                            {
                                "Provider": "com.apple.wallpaper.choice.aerials",
                                "Files": [],
                                "Configuration": b"",
                            }
                        ],
                        "Shuffle": {
                            "Type": "afterDuration",
                            "Duration": [2341, 16172123445939666944],
                        },
                    },
                },
            },
        }
        plistlib.dump(new_plist, f)
        f.close()
    subprocess.call(["killall", "WallpaperAgent"])


def set_win_wallpaper(img):
    """Set the wallpaper on Windows."""
    # There's a different command depending on the architecture
    # of Windows. We check the PROGRAMFILES envar since using
    # platform is unreliable.
    if "x86" in os.environ["PROGRAMFILES"]:
        ctypes.windll.user32.SystemParametersInfoW(20, 0, img, 3)
    else:
        # 'W' functions take unicode strings,
        # while 'A' functions take UTF-8 bytestrings.
        # (Python 3 strings are Unicode by default.)
        ctypes.windll.user32.SystemParametersInfoA(20, 0, str.encode(img), 3)


def change(img):
    """Set the wallpaper."""
    if not os.path.isfile(img):
        return

    desktop = get_desktop_env()

    if OS == "Darwin":
        set_mac_wallpaper(img)

    elif OS == "Windows":
        set_win_wallpaper(img)

    else:
        set_desktop_wallpaper(desktop, img)

    logging.info("Set the new wallpaper.")


def get(cache_dir=CACHE_DIR):
    """Get the current wallpaper."""
    current_wall = os.path.join(cache_dir, "wal")

    if os.path.isfile(current_wall):
        # make sure the file has some content in it,
        contents = util.read_file(current_wall)

        if len(contents) > 0:
            return contents[0]

    return "None"



"""
Generate a colorscheme using ColorThief.
"""

import logging
import sys

try:
    from colorthief import ColorThief

except ImportError:
    logging.error("ColorThief wasn't found on your system.")
    logging.error("Try another backend. (wal --backend)")
    sys.exit(1)

from .. import colors
from .. import util


def gen_colors(img):
    """Loop until 16 colors are generated."""
    color_cmd = ColorThief(img).get_palette

    for i in range(0, 10, 1):
        raw_colors = color_cmd(color_count=8 + i)

        if len(raw_colors) >= 8:
            break

        if i == 10:
            logging.error("ColorThief couldn't generate a suitable palette.")
            sys.exit(1)

        else:
            logging.warning("ColorThief couldn't generate a palette.")
            logging.warning("Trying a larger palette size %s", 8 + i)

    return [util.rgb_to_hex(color) for color in raw_colors]


def adjust(cols, light, **kwargs):
    """Create palette.
    :keyword-args:
    -    c16: use 16 colors through specified method - [ "lighten" | "darken" ]
    """
    if "c16" in kwargs:
        cols16 = kwargs["c16"]
    else:
        cols16 = False
    cols.sort(key=util.rgb_to_yiq)
    raw_colors = [*cols, *cols]
    for color in raw_colors:
        color = util.lighten_color(color, 0.40)
    raw_colors[0] = util.darken_color(cols[0], 0.80)

    return colors.generic_adjust(raw_colors, light, c16=cols16)


def get(img, light=False, **kwargs):
    """Get colorscheme.
    :keyword-args:
    -    c16: use 16 colors through specified method - [ "lighten" | "darken" ]
    """
    if "c16" in kwargs:
        cols16 = kwargs["c16"]
    else:
        cols16 = False
    cols = gen_colors(img)
    return adjust(cols, light, c16=cols16)



"""
                                      '||
... ...  .... ... ... ... ...  ....    ||
 ||'  ||  '|.  |   ||  ||  |  '' .||   ||
 ||    |   '|.|     ||| |||   .|' ||   ||
 ||...'     '|       |   |    '|..'|' .||.
 ||      .. |
''''      ''
Created by Dylan Araps.
"""

import argparse
import logging
import os
import shutil
import sys

from .settings import __version__, CACHE_DIR, CONF_DIR
from . import colors
from . import export
from . import image
from . import reload
from . import sequences
from . import theme
from . import util
from . import wallpaper
from . import donation
from . import eastereggs


show_colorama_warning = False
if sys.platform.startswith("win"):
    try:
        import colorama
        colorama.just_fix_windows_console()
    except ImportError:
        no_colorama_terms = [
                "wezterm",
                "alacritty",
                "hyper",
                "putty",
                "ghostty",
                "mintty",
                ]
        if os.environ.get("TERMINAL") not in no_colorama_terms and \
           os.environ.get("TERM_PROGRAM") not in no_colorama_terms:
            show_colorama_warning = True


def get_args():
    """Get the script arguments."""
    description = "wal - Generate colorschemes on the fly"
    arg = argparse.ArgumentParser(description=description)

    arg.add_argument(
        "-a",
        metavar='"alpha"',
        help="Set terminal background transparency. \
              Must be a number between 0 and 100. \
              *Only works in terminals that implement OSC-11 (URxvt)*",
    )

    arg.add_argument(
        "-b", metavar="background", help="Custom background color to use."
    )

    arg.add_argument(
        "--fg", metavar="foreground", help="Custom foreground color to use."
    )

    arg.add_argument(
        "--backend",
        metavar="backend",
        help="Which color backend to use. \
                           Use 'wal --backend' to list backends.",
        const="list_backends",
        type=str,
        nargs="?",
    )

    arg.add_argument(
        "--out-dir",
        metavar="out_dir",
        help="Cache dir to export themes. \
              Default is 'XDG_CACHE_HOME/wal'. \
              This can also be set with the env var: 'PYWAL_CACHE_DIR'",
        type=str,
        nargs="?",
    )

    arg.add_argument(
        "--theme",
        "-f",
        metavar="/path/to/file or theme_name",
        help="Which colorscheme file to use. \
                           Use 'wal --theme' to list builtin and user themes.",
        const="list_themes",
        nargs="?",
    )

    arg.add_argument(
        "--iterative",
        action="store_true",
        help="When pywal is given a directory as input and this "
        "flag is used: Go through the images in order "
        "instead of shuffled.",
    )

    arg.add_argument(
        "--cols16",
        metavar="method",
        required=False,
        nargs="?",
        default=False,
        const="darken",
        choices=["darken", "lighten"],
        help="Use 16 color output " '"darken" or "lighten" default: darken',
    )

    arg.add_argument(
        "--recursive",
        action="store_true",
        help="When pywal is given a directory as input and this "
        "flag is used: Search for images recursively in "
        "subdirectories instead of the root only.",
    )

    arg.add_argument(
        "--saturate", metavar="0.0-1.0", help="Set the color saturation."
    )

    arg.add_argument(
        "--preview",
        action="store_true",
        help="Print the current color palette.",
    )

    arg.add_argument(
        "--vte",
        action="store_true",
        help="Fix text-artifacts printed in VTE terminals.",
    )

    arg.add_argument(
        "-c", action="store_true", help="Delete all cached colorschemes."
    )

    arg.add_argument(
        "-i",
        metavar='"/path/to/img.jpg"',
        help="Which image or directory to use.",
    )

    arg.add_argument(
        "-l", action="store_true", help="Generate a light colorscheme."
    )

    arg.add_argument(
        "-n", action="store_true", help="Skip setting the wallpaper."
    )

    arg.add_argument(
        "-o",
        metavar='"script_name"',
        action="append",
        help='External script to run after "wal".',
    )

    arg.add_argument(
        "-p",
        metavar='"theme_name"',
        help="permanently save theme to "
        "$XDG_CONFIG_HOME/wal/colorschemes with "
        "the specified name",
    )

    arg.add_argument(
        "-q", action="store_true", help="Quiet mode, don't print anything."
    )

    arg.add_argument(
        "-r",
        action="store_true",
        help="'wal -r' is deprecated: Use \
                           (cat ~/.cache/wal/sequences &) instead.",
    )

    arg.add_argument(
        "-R", action="store_true", help="Restore previous colorscheme."
    )

    arg.add_argument(
        "-s", action="store_true", help="Skip changing colors in terminals."
    )

    arg.add_argument(
        "-t", action="store_true", help="Skip changing colors in tty."
    )

    arg.add_argument("-v", action="store_true", help='Print "wal" version.')

    arg.add_argument(
        "-w",
        action="store_true",
        help="Use last used wallpaper for color generation.",
    )

    arg.add_argument(
        "-e",
        action="store_true",
        help="Skip reloading gtk/xrdb/i3/sway/polybar",
    )

    arg.add_argument(
        "--contrast",
        metavar="1.0-21.0",
        required=False,
        type=float,
        nargs="?",
        help="Specify a minimum contrast ratio between palette "
        "colors and the source image according to W3 "
        "contrast specifications. Values between 1.5-4.5 "
        "typically work best.",
    )

    return arg


def parse_args_exit(parser):
    """Process args that exit."""
    args = parser.parse_args()

    if len(sys.argv) <= 1:
        parser.print_help()
        sys.exit(1)

    if args.v:
        parser.exit(0, "wal %s\n" % __version__)

    if args.preview:
        print("Current colorscheme:", sep="")
        colors.palette()
        sys.exit(0)

    if args.r:
        reload.colors()
        sys.exit(0)

    if args.c:
        scheme_dir = os.path.join(CACHE_DIR, "schemes")
        shutil.rmtree(scheme_dir, ignore_errors=True)
        sys.exit(0)

    if (
        not args.i
        and not args.theme
        and not args.R
        and not args.w
        and not args.backend
    ):
        parser.error(
            "No input specified.\n" "--backend, --theme, -i or -R are required."
        )

    if args.theme == "list_themes":
        theme.list_out()
        sys.exit(0)

    if args.backend == "list_backends":
        print(
            "\n - ".join(
                ["\033[1;32mBackends\033[0m:", *colors.list_backends()]
            )
        )
        sys.exit(0)


def parse_args(parser):
    """Process args."""
    args = parser.parse_args()

    if args.q:
        logging.getLogger().disabled = True
        sys.stdout = sys.stderr = open(os.devnull, "w")

    if args.a:
        util.alpha_integrify(args.a)
        util.Color.passed_alpha_num = args.a
        util.Color.alpha_num = args.a or util.Color.alpha_num

    if args.i and not args.theme:
        image_file = image.get(
            args.i, iterative=args.iterative, recursive=args.recursive
        )
        colors_plain = colors.get(
            image_file,
            args.l,
            args.backend,
            sat=args.saturate,
            c16=args.cols16,
            cst=args.contrast,
        )

    if args.theme:
        colors_plain = theme.file(args.theme, args.l, c16=args.cols16)
        if args.i:
            colors_plain["wallpaper"] = args.i

    if args.R:
        colors_plain = theme.file(os.path.join(CACHE_DIR, "colors.json"))

    if args.w:
        cached_wallpaper = util.read_file(os.path.join(CACHE_DIR, "wal"))
        colors_plain = colors.get(
            cached_wallpaper[0],
            args.l,
            args.backend,
            sat=args.saturate,
            c16=args.cols16,
            cst=args.contrast,
        )

    if args.b:
        args.b = "#%s" % (args.b.strip("#"))
        colors_plain["special"]["background"] = args.b
        colors_plain["colors"]["color0"] = args.b

    if args.fg:
        args.fg = "#%s" % (args.fg.strip("#"))
        colors_plain["special"]["foreground"] = args.fg
        colors_plain["colors"]["color15"] = args.fg

    if not args.n:
        wallpaper.change(colors_plain["wallpaper"])

    if args.p:
        theme.save(colors_plain, args.p, args.l)

    sequences.send(colors_plain, to_send=not args.s, vte_fix=args.vte)

    if sys.stdout.isatty():
        colors.palette()

    if args.out_dir:
        export.every(colors_plain, CACHE_DIR)
    else:
        export.every(colors_plain)

    if not args.e:
        reload.env(tty_reload=not args.t)

    if args.o:
        for cmd in args.o:
            util.disown([cmd])


def main():
    """Main script function."""
    global CACHE_DIR
    default_cache_dir = CACHE_DIR
    util.create_dir(os.path.join(CONF_DIR, "templates"))
    util.create_dir(os.path.join(CONF_DIR, "colorschemes/light/"))
    util.create_dir(os.path.join(CONF_DIR, "colorschemes/dark/"))

    util.setup_logging()

    if show_colorama_warning:
        logging.warning("colorama is not present")

    parser = get_args()

    args = parser.parse_args()
    if args.out_dir:
        CACHE_DIR = args.out_dir
    else:
        CACHE_DIR = default_cache_dir

    parse_args_exit(parser)
    parse_args(parser)
    donation.donation_message()
    eastereggs.eemsg()


if __name__ == "__main__":
    main()



"""
Generate a colorscheme using Schemer2.
"""

import logging
import shutil
import subprocess
import sys

from .. import colors
from .. import util


def gen_colors(img):
    """Generate a colorscheme using Colorz."""
    cmd = ["schemer2", "-format", "img::colors", "-minBright", "75", "-in"]
    return subprocess.check_output([*cmd, img]).splitlines()


def adjust(cols, light, **kwargs):
    """Create palette.
    :keyword-args:
    -    c16: use 16 colors through specified method - [ "lighten" | "darken" ]
    """
    if "c16" in kwargs:
        cols16 = kwargs["c16"]
    else:
        cols16 = False
    cols.sort(key=util.rgb_to_yiq)
    raw_colors = [*cols[8:], *cols[8:]]
    raw_colors[0] = util.darken_color(cols[0], 0.80)

    return colors.generic_adjust(raw_colors, light, c16=cols16)


def get(img, light=False, **kwargs):
    """Get colorscheme.
    :keyword-args:
    -    c16: use 16 colors through specified method - [ "lighten" | "darken" ]
    """
    if "c16" in kwargs:
        cols16 = kwargs["c16"]
    else:
        cols16 = False
    if not shutil.which("schemer2"):
        logging.error("Schemer2 wasn't found on your system.")
        logging.error("Try another backend. (wal --backend)")
        sys.exit(1)

    cols = [col.decode("UTF-8") for col in gen_colors(img)]
    return adjust(cols, light, c16=cols16)



"""
Send sequences to all open terminals.
"""

import glob
import logging
import os
import subprocess

from .settings import CACHE_DIR, OS
from . import util


def set_special(index, color, iterm_name="h", alpha=100):
    """Convert a hex color to a special sequence."""
    if OS == "Darwin" and iterm_name:
        return "\033]P%s%s\033\\" % (iterm_name, color.strip("#"))

    if index in [11, 708] and alpha != "100":
        return "\033]%s;[%s]%s\033\\" % (index, alpha, color)

    return "\033]%s;%s\033\\" % (index, color)


def set_color(index, color):
    """Convert a hex color to a text color sequence."""
    if OS == "Darwin" and index < 20:
        return "\033]P%1x%s\033\\" % (index, color.strip("#"))

    return "\033]4;%s;%s\033\\" % (index, color)


def set_iterm_tab_color(color):
    """Set iTerm2 tab/window color"""
    return (
        "\033]6;1;bg;red;brightness;%s\a"
        "\033]6;1;bg;green;brightness;%s\a"
        "\033]6;1;bg;blue;brightness;%s\a"
    ) % (*util.hex_to_rgb(color),)


def create_sequences(colors, vte_fix=False):
    """Create the escape sequences."""
    alpha = colors["alpha"]

    # Colors 0-15.
    sequences = [
        set_color(index, colors["colors"]["color%s" % index])
        for index in range(16)
    ]

    # Special colors.
    # Source: https://goo.gl/KcoQgP
    # 10 = foreground, 11 = background, 12 = cursor foreground
    # 13 = mouse foreground, 708 = background border color.
    sequences.extend(
        [
            set_special(10, colors["special"]["foreground"], "g"),
            set_special(11, colors["special"]["background"], "h", alpha),
            set_special(12, colors["special"]["cursor"], "l"),
            set_special(13, colors["special"]["foreground"], "j"),
            set_special(17, colors["special"]["foreground"], "k"),
            set_special(19, colors["special"]["background"], "m"),
            set_color(232, colors["special"]["background"]),
            set_color(256, colors["special"]["foreground"]),
            set_color(257, colors["special"]["background"]),
        ]
    )

    if not vte_fix:
        sequences.extend(
            set_special(708, colors["special"]["background"], "", alpha)
        )

    if OS == "Darwin":
        sequences += set_iterm_tab_color(colors["special"]["background"])

    return "".join(sequences)


def send(colors, cache_dir=CACHE_DIR, to_send=True, vte_fix=False):
    """Send colors to all open terminals."""
    if OS == "Darwin":
        devices = glob.glob("/dev/ttys00[0-9]*")
    elif OS == "OpenBSD":
        devices = subprocess.check_output(
            "ps -o tty | sed -e 1d -e s#^#/dev/# | sort | uniq",
            shell=True,
            universal_newlines=True,
        ).split()
    else:
        devices = glob.glob("/dev/pts/[0-9]*")

    sequences = create_sequences(colors, vte_fix)

    if not util.has_fcntl:
        logging.warning(util.fcntl_warning)

    # Send data to open terminal devices.
    if to_send:
        for dev in devices:
            if dev == "/dev/pts/0":
                if os.environ.get("DESKTOP_SESSION") == "plasma":
                    continue
            util.save_file(sequences, dev)

    util.save_file(sequences, os.path.join(cache_dir, "sequences"))
    logging.info("Set terminal colors.")



"""
Generate a colorscheme using imagemagick.
"""

import logging
import re
import shutil
import subprocess
import sys

from .. import colors
from .. import util


def imagemagick(color_count, img, magick_command):
    """Call Imagemagick to generate a scheme."""
    flags = [
        "-resize",
        "25%",
        "-colors",
        str(color_count),
        "-unique-colors",
        "txt:-",
    ]
    img += "[0]"

    try:
        output = subprocess.check_output(
            [*magick_command, img, *flags], stderr=subprocess.STDOUT
        ).splitlines()
    except subprocess.CalledProcessError as Err:
        logging.error("Imagemagick error: %s", Err)
        logging.error(
            "IM 7 disables stdout by default, check the wiki for the fix."
        )
        sys.exit(1)
    return output


def has_im():
    """Check to see if the user has im installed."""
    if shutil.which("magick"):
        return ["magick", "convert"]

    if shutil.which("convert"):
        return ["convert"]

    logging.error("Imagemagick wasn't found on your system.")
    logging.error("Try another backend. (wal --backend)")
    sys.exit(1)


def try_gen_in_range(img, magick_command):
    for i in range(0, 20, 1):
        raw_colors = imagemagick(16 + i, img, magick_command)

        if len(raw_colors) > 16:
            break

        if i == 19:
            logging.error("Imagemagick couldn't generate a suitable palette.")
            sys.exit(1)

        else:
            logging.warning("Imagemagick couldn't generate a palette.")
            logging.warning("Trying a larger palette size %s", 16 + i)
    return raw_colors


def gen_colors(img):
    """Format the output from imagemagick into a list
    of hex colors."""
    magick_command = has_im()

    raw_colors = try_gen_in_range(img, magick_command)

    try:
        out = [re.search("#.{6}", str(col)).group(0) for col in raw_colors[1:]]
    except AttributeError:
        if magick_command == ["magick", "convert"]:
            logging.warning("magick convert failed, using only magick")
            magick_command = ["magick"]
            raw_colors = try_gen_in_range(img, magick_command)
            out = [
                re.search("#.{6}", str(col)).group(0) for col in raw_colors[1:]
            ]

    return out


def adjust(cols, light, **kwargs):
    """Adjust the generated colors and store them in a dict that
    we will later save in json format.
    :keyword-args:
    -    c16: use 16 colors through specified method - [ "lighten" | "darken" ]
    """
    if "c16" in kwargs:
        cols16 = kwargs["c16"]
    else:
        cols16 = False
    raw_colors = cols[:1] + cols[8:16] + cols[8:-1]

    return colors.generic_adjust(raw_colors, light, c16=cols16)


def get(img, light=False, **kwargs):
    """Get colorscheme.
    :keyword-args:
    -    c16: use 16 colors through specified method - [ "lighten" | "darken" ]
    """
    if "c16" in kwargs:
        cols16 = kwargs["c16"]
    else:
        cols16 = False
    colors = gen_colors(img)
    # it is possible we could have picked garbage data
    garbage = "# Image"
    if garbage in colors:
        colors.remove(garbage)
    return adjust(colors, light, c16=cols16)



import datetime
import logging


def donation_message():
    url = "https://github.com/eylles/pywal16"
    donate_msg = "if you enjoy pywal16 consider donating at: {}".format(url)
    if datetime.datetime.now().month == 12:
        logging.info(donate_msg)
    else:
        msec_mod = datetime.datetime.now().microsecond % 20
        if msec_mod >= 8:
            logging.info(donate_msg)



"""wal - setup.py"""

import sys
import setuptools

try:
    import pywal
except ImportError:
    print("error: pywal requires Python 3.5 or greater.")
    sys.exit(1)

is_win = sys.platform.startswith("win")

LONG_DESC = open("README.md").read()
VERSION = pywal.__version__
DOWNLOAD = "https://github.com/eylles/pywal16/archive/%s.tar.gz" % VERSION

install_requires = []
if is_win:
    install_requires += ["colorama"]

setuptools.setup(
    name="pywal16",
    version=VERSION,
    author="Dylan Araps",
    author_email="dylan.araps@gmail.com",
    description="Generate and change color-schemes on the fly",
    long_description_content_type="text/markdown",
    long_description=LONG_DESC,
    keywords="wal colorscheme terminal-emulators changing-colorschemes",
    license="MIT",
    url="https://github.com/eylles/pywal16",
    download_url=DOWNLOAD,
    classifiers=[
        "Environment :: X11 Applications",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    data_files=[('man/man1', ['data/man/man1/wal.1'])],
    packages=["pywal"],
    entry_points={"console_scripts": ["wal=pywal.__main__:main"]},
    python_requires=">=3.5",
    install_requires=install_requires,
    extras_require={
        "colorthief": [
            "colorthief",
        ],  # known to work with 0.2.1
        "colorz": [
            "colorz",
        ],  # NOTE heavy, scipy dependency
        "fast-colorthief": [
            "fast-colorthief",
        ],
        "haishoku": [
            "haishoku",
        ],
        "modern_colorthief": [
            "modern_colorthief",
        ],
        "all": [
            "colorthief",
            "colorz",
            "fast-colorthief",
            "haishoku",
            "modern_colorthief",
        ],  # convience, all of the above
    },
    include_package_data=True,
    zip_safe=False,
)



"""
Misc helper functions.
"""

import colorsys
import json
import logging
import os
import platform
import re
import shutil
import subprocess
import sys
import hashlib
import copy

has_fcntl = False
fcntl_warning = ""

try:
    import fcntl

    has_fcntl = True
except ImportError:
    fcntl_warning = "{}, {}".format(
        "can't skip blocking io in current platform",
        "program could hang indefinitely",
    )


class Color:
    """Color formats."""

    alpha_num = "100"
    passed_alpha_num = None

    def __init__(self, hex_color):
        self.hex_color = hex_color

    def __str__(self):
        return self.hex_color

    @property
    def rgb(self):
        """Convert a hex color to rgb."""
        return "%s,%s,%s" % (*hex_to_rgb(self.hex_color),)

    @property
    def rgbspace(self):
        """Convert a hex color to rgb separated by spaces."""
        return "%s %s %s" % (*hex_to_rgb(self.hex_color),)

    @property
    def xrgba(self):
        """Convert a hex color to xrdb rgba."""
        return hex_to_xrgba(self.hex_color)

    @property
    def rgba(self):
        """Convert a hex color to rgba."""
        return "rgba(%s,%s,%s,%s)" % (
            *hex_to_rgb(self.hex_color),
            self.alpha_dec,
        )

    @property
    def hex_argb(self):
        """Convert an alpha hex color to argb hex."""
        al_val = alpha_integrify(self.alpha_num)
        return "#%02X%s" % (
            int(int(al_val) * 255 / 100),
            self.hex_color[1:],
        )

    @property
    def alpha(self):
        """Add URxvt alpha value to color."""
        al_val = alpha_integrify(self.alpha_num)
        return "[%s]%s" % (al_val, self.hex_color)

    @property
    def alpha_dec(self):
        """Export the alpha value as a decimal number in [0, 1]."""
        al_val = alpha_integrify(self.alpha_num)
        return int(al_val) / 100

    @property
    def alpha_hex(self):
        """Export the alpha value as a hexdecimal number in [00, FF]."""
        al_val = alpha_integrify(self.alpha_num)
        return "%02X" % (int(int(al_val) * 255 / 100))

    @property
    def decimal(self):
        """Export color in decimal."""
        return "%s%s" % ("#", int(self.hex_color[1:], 16))

    @property
    def decimal_strip(self):
        """Strip '#' from decimal color."""
        return int(self.hex_color[1:], 16)

    @property
    def octal(self):
        """Export color in octal."""
        return "%s%s" % ("#", oct(int(self.hex_color[1:], 16))[2:])

    @property
    def octal_strip(self):
        """Strip '#' from octal color."""
        return oct(int(self.hex_color[1:], 16))[2:]

    @property
    def strip(self):
        """Strip '#' from color."""
        return self.hex_color[1:]

    @property
    def red(self):
        """Red value as float between 0 and 1."""
        return "%.3f" % (hex_to_rgb(self.hex_color)[0] / 255.0)

    @property
    def green(self):
        """Green value as float between 0 and 1."""
        return "%.3f" % (hex_to_rgb(self.hex_color)[1] / 255.0)

    @property
    def blue(self):
        """Blue value as float between 0 and 1."""
        return "%.3f" % (hex_to_rgb(self.hex_color)[2] / 255.0)

    @property
    def red_hex(self):
        """Red value as hex."""
        return "%s" % (self.hex_color)[1:3]

    @property
    def green_hex(self):
        """Green value as hex."""
        return "%s" % (self.hex_color)[3:5]

    @property
    def blue_hex(self):
        """Blue value as hex."""
        return "%s" % (self.hex_color)[5:]

    @property
    def red_dec(self):
        """Red value as decimal."""
        return "%s" % hex_to_rgb(self.hex_color)[0]

    @property
    def green_dec(self):
        """Green value as decimal."""
        return "%s" % hex_to_rgb(self.hex_color)[1]

    @property
    def blue_dec(self):
        """Blue value as decimal."""
        return "%s" % hex_to_rgb(self.hex_color)[2]

    @property
    def w3_luminance(self):
        """Luminance value of the color according to W3 formula"""

        color_channels = [float(self.red), float(self.green), float(self.blue)]
        for index, channel in enumerate(color_channels):
            if channel <= 0.04045:
                color_channels[index] = channel / 12.92
            else:
                color_channels[index] = ((channel + 0.055) / 1.055) ** 2.4

        return (
            (0.2126 * color_channels[0])
            + (0.7152 * color_channels[1])
            + (0.0722 * color_channels[2])
        )

    def lighten(self, percent):
        """Lighten color by percent."""
        percent = float(re.sub(r"[\D\.]", "", str(percent)))
        return Color(lighten_color(self.hex_color, percent / 100))

    def darken(self, percent):
        """Darken color by percent."""
        percent = float(re.sub(r"[\D\.]", "", str(percent)))
        return Color(darken_color(self.hex_color, percent / 100))

    def saturate(self, percent):
        """Saturate a color."""
        percent = float(re.sub(r"[\D\.]", "", str(percent)))
        return Color(saturate_color(self.hex_color, percent / 100))

    def adjust_alpha(self, alpha="100"):
        adjusted = copy.copy(self)
        adjusted.alpha_num = alpha
        return adjusted


def read_file(input_file):
    """Read data from a file and trim newlines."""
    with open(input_file, "r") as file:
        return file.read().splitlines()


def read_file_json(input_file):
    """Read data from a json file."""
    with open(input_file, "r") as json_file:
        return json.load(json_file)


def read_file_raw(input_file):
    """Read data from a file as is, don't strip
    newlines or other special characters."""
    with open(input_file, "r") as file:
        return file.readlines()


def save_file(data, export_file):
    """Write data to a file."""
    create_dir(os.path.dirname(export_file))

    if has_fcntl:
        try:
            with open(export_file, "w") as file:
                # Get the current flags and add non-blocking mode
                # to skip TTYs suspended by Flow Control
                # https://www.gnu.org/software/libc/manual/html_node/Getting-File-Status-Flags.html
                # https://www.gnu.org/software/libc/manual/html_node/Open_002dtime-Flags.html
                flags = fcntl.fcntl(file, fcntl.F_GETFL)
                fcntl.fcntl(file, fcntl.F_SETFL, flags | os.O_NONBLOCK)
                file.write(data)
        except PermissionError:
            logging.warning("Couldn't write to %s.", export_file)
        except BlockingIOError:
            logging.warning(
                "Couldn't write to %s, not accepting data", export_file
            )
    else:
        try:
            with open(export_file, "w") as file:
                file.write(data)
        except PermissionError:
            logging.warning("Couldn't write to %s.", export_file)


def save_file_json(data, export_file):
    """Write data to a json file."""
    create_dir(os.path.dirname(export_file))

    with open(export_file, "w") as file:
        json.dump(data, file, indent=4)


def get_img_checksum(img):
    checksum = hashlib.new("md5", usedforsecurity=False)
    with open(img, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            checksum.update(chunk)
    return checksum.hexdigest()


def create_dir(directory):
    """Alias to create the cache dir."""
    os.makedirs(directory, exist_ok=True)


def setup_logging():
    """Logging config."""
    logging.basicConfig(
        format=(
            "[%(levelname)s\033[0m] "
            "\033[1;31m%(module)s\033[0m: "
            "%(message)s"
        ),
        level=logging.INFO,
        stream=sys.stdout,
    )
    logging.addLevelName(logging.ERROR, "\033[1;31mE")
    logging.addLevelName(logging.INFO, "\033[1;32mI")
    logging.addLevelName(logging.WARNING, "\033[1;33mW")


def hex_to_rgb(color):
    """Convert a hex color to rgb."""
    return tuple(bytes.fromhex(color.strip("#")))


def hex_to_xrgba(color):
    """Convert a hex color to xrdb rgba."""
    col = color.lower().strip("#")
    return "%s%s/%s%s/%s%s/ff" % (*col,)


def rgb_to_hex(color):
    """Convert an rgb color to hex."""
    return "#%02x%02x%02x" % (*color,)


def darken_color(color, amount):
    """Darken a hex color."""
    color = [int(col * (1 - amount)) for col in hex_to_rgb(color)]
    return rgb_to_hex(color)


def lighten_color(color, amount):
    """Lighten a hex color."""
    color = [int(col + (255 - col) * amount) for col in hex_to_rgb(color)]
    return rgb_to_hex(color)


def alpha_integrify(alpha_value):
    """
    ensure the alpha string is an int between 0 and 100

    return: int string between 0 an 100
    """
    # could be a string containing a float like 0.7
    a = float(alpha_value)
    if a < 0:
        a = abs(a)
    if a < 1:
        a = a * 100
    if a > 100:
        a = 100
    a = int(a)
    a = str(a)
    return a


def blend_color(color, color2):
    """Blend two colors together."""
    r1, g1, b1 = hex_to_rgb(color)
    r2, g2, b2 = hex_to_rgb(color2)

    r3 = int(0.5 * r1 + 0.5 * r2)
    g3 = int(0.5 * g1 + 0.5 * g2)
    b3 = int(0.5 * b1 + 0.5 * b2)

    return rgb_to_hex((r3, g3, b3))


def saturate_color(color, amount):
    """Saturate a hex color."""
    r, g, b = hex_to_rgb(color)
    r, g, b = [x / 255.0 for x in (r, g, b)]
    h, l, s = colorsys.rgb_to_hls(r, g, b)
    s = amount
    r, g, b = colorsys.hls_to_rgb(h, l, s)
    r, g, b = [x * 255.0 for x in (r, g, b)]

    return rgb_to_hex((int(r), int(g), int(b)))


def rgb_to_yiq(color):
    """Sort a list of colors."""
    return colorsys.rgb_to_yiq(*hex_to_rgb(color))


def disown(cmd):
    """Call a system command in the background,
    disown it and hide it's output."""
    subprocess.Popen(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)


def get_pid(name):
    """Check if process is running by name."""
    if not shutil.which("pidof"):
        return False

    try:
        if platform.system() != "Darwin":
            subprocess.check_output(["pidof", "-s", name])
        else:
            subprocess.check_output(["pidof", name])

    except subprocess.CalledProcessError:
        return False

    return True


def has_im():
    """Check to see if the user has im installed."""
    if shutil.which("magick"):
        return "magick"

    if shutil.which("convert"):
        return "convert"

    logging.error("Problem running image averaging command.")
    logging.error("Imagemagick wasn't found on your system.")
    sys.exit(1)


def image_average_color(img):
    """Get the average color of an image using imagemagick
    by resizing to 1x1"""
    # Attempt to run the imagemagick command
    # Resizes to 1x1 and enumerates all pixel data (one pixel) to stdout
    # Command adapted from a stackoverflow thread, but tinkered with because the
    # thread was a decade old:
    # # https://stackoverflow.com/questions/25488338/how-to-find-average-color-of-an-image-with-imagemagick
    cmd_flags = [
        "-resize",
        "1x1!",
        "-format",
        '"%[fx:int(255*r+.5)],%[fx:int(255*g+.5)],%[fx:int(255*b+.5)]"',
        "txt:-",
    ]
    magick_command = has_im()
    try:
        magick_output = subprocess.run(
            [magick_command, img] + cmd_flags, stdout=subprocess.PIPE
        )
    except subprocess.CalledProcessError as Err:
        logging.error(
            "Problem running image averaging command. Is imagemagick installed?"
        )
        logging.error("Imagemagick error: %s", Err)
        return ""

    # Regex hex code from the command output
    return re.search("#[0-9A-Fa-f]{6}", magick_output.stdout.decode("utf-8"))[0]



import datetime

are_we_wayland_yet = """
You will never be a real display server. You have no hardware cursors,
you have no xrandr, you have no setxkbmap. You are a toy project twisted
by Red Hat and GNOME into a crude mockery of X11’s perfection.

All the “validation” you get is two-faced and half-hearted. Behind
your back people mock you. Your developers are disgusted and ashamed of you,
your “users” laugh at your lack of features behind closed doors.

Linux users are utterly repulsed by you. Thousands of years of evolution have
allowed them to sniff out defective software with incredible efficiency.
Even Wayland sessions that “work” look uncanny and unnatural to a seasoned
sysadmin. Your bizarre render loop is a dead giveaway. And even if you manage
to get a drunk Arch user home with you, he’ll turn tail and bolt the second he
gets a whiff of your high latency due to forced VSync.

You will never be happy. You wrench out a fake smile every single morning and
tell yourself it’s going to be ok, but deep inside you feel the technical
debt creeping up like a weed, ready to crush you under the unbearable weight.

Eventually it’ll be too much to bear - you’ll log into the GitLab instance,
select the project, press Delete, and plunge it into the cold abyss. Your
users will find the deletion notice, heartbroken but relieved that they no
longer have to live with the unbearable shame and disappointment. They’ll
remember you as the biggest failure of open source development, and every
passerby for the rest of eternity will know a badly run project has failed
there. Your code will decay and go to historical archives, and all that will
remain of your legacy is a codebase that is unmistakably poorly written.

This is your fate. This is what you chose. There is no turning back
"""

we_will_replace_x11 = """
To you who stood in the X.Org Developer's Conference and waved banners of
obsolence and chanted "You will not replace us," please know one thing:

We will replace you.

Despite your insistence to the contrary, know that you have sown the seeds of
your own irrelevance. This is not to say that X.Org Developers will be
marginalized and replaced. But rather that people who believe in superiority
based on programming in C and C++ have no place in our modern society.

Your ahistorical, empty idolatry to protocol "purity" will be consigned to the
trash heap of history.

You will be replaced by Wayland Developers such as myself. You will be replaced
by the many fine people I work with and am friends with, people who come in all
protocols, compositors and believe in every programming paradigm imaginable.
You will be replaced by people like my mother, a former X.Org developer now
transitioned into a Wayland protocol developer who abhors everything your
display server and windowing system stand for.

You will be replaced by people who believe in the Wayland project, in protocol
proliferation, in compositor implementation, who reject your obsolete ways and
choose to embrace the idea that there should be A unix desktop that does not
depend on technology developed in the 80s, in 40 years no one will be a X.Org
user as all your descendants will be Wayland users.
"""

ricing_feels_soulless = """
Holy shit I just realized why ricing feels so soulless in the age of hyprland,
it's because nobody respects color schemes anymore.

Gone are the days of base16 or even base8, now everything is pywal (or whatever
derivative is the current meta) with poorly recolored wallpapers. People don't
care about color balance or variety anymore, that's why everything looks so
samey and why it feels like there's a "hyprland aesthetic" -- it's
transparency+blur with a wallpaper and colorscheme that have no variety;
they're almost entirely monochrome or otherwise poorly created by a program
rather than someone with an actual eye.

On top of this, I think another major part is how people don't respect pairing
wallpapers and color schemes anymore. It's always premade wallpapers to match
the theme (catppuccin/gruvbox), wallpapers recolored by a program to match the
theme (aster or equivalent), or themes created by a program to match the
wallpaper (pywal or equivalent). There used to be an art to this, but people
don't care about it anymore.

And even when people don't use pywal, it's going to be gruvbox or catppuccin or
whatever other soulless theme is the meta. I remember hating nord when it was
popular, but I would give anything to see some nord rices instead of the same
catppuccin and gruvbox ones.

TLDR people don't care about color schemes anymore and would prefer to use a
poorly cobbled together one from a program or a premade theme rather than
putting any form of effort in.
"""

gimp = """
GIMP = Green Is My Pepper
"""

guido = """
I'm not too worried about the associations with mob assassins
that some people have. :-)
"""

pywal_is_so_cute = """
Pywal is so cute omg(⁄ ⁄•⁄ω⁄•⁄ ⁄)⁄ when you hold it on your hand and
it starts working its like its nuzzling you (/ω＼) or when it perks up and
look at you like" owo nya? :3c" hehe ~ pywal-kun is happy to see me!!（＾ワ＾）
and the most adorable thing ever is when colorscheme-sama comes out but theyre
rlly shy so u have to work hard!!(๑•̀ㅁ•́๑)✧ but when pywal-kun and
colorscheme-sama meet and theyre blushing and all like "uwaaa~!" (ﾉ´ヮ´)ﾉ*: ･ﾟ
hehehe~  pywal-kun is so adorable (●´Д｀●)・:*:・
"""

the_finest_in_the_district = """
Python, It's the finest in the district!
"""


def eemsg():
    if datetime.datetime.now().month == 1 and datetime.datetime.now().day == 31:
        print(guido)
    if datetime.datetime.now().month == 4 and datetime.datetime.now().day == 1:
        print(ricing_feels_soulless)
    if datetime.datetime.now().month == 3 and datetime.datetime.now().day == 16:
        print(gimp)
    if datetime.datetime.now().month == 6 and datetime.datetime.now().day == 21:
        print(pywal_is_so_cute)
    if datetime.datetime.now().month == 9 and datetime.datetime.now().day == 15:
        print(we_will_replace_x11)
    if datetime.datetime.now().month == 9 and datetime.datetime.now().day == 30:
        print(are_we_wayland_yet)

    if datetime.datetime.now().month == 12 and datetime.datetime.now().day == 3:
        print(the_finest_in_the_district)



"""
Generate a palette using various backends.
"""

import colorsys
import logging
import os
import random
import re
import sys

from . import theme
from . import util
from .settings import CACHE_DIR, MODULE_DIR, __cache_version__


def list_backends():
    """List color backends."""
    return [
        b.name.replace(".py", "")
        for b in os.scandir(os.path.join(MODULE_DIR, "backends"))
        if "__" not in b.name
    ]


def normalize_img_path(img: str):
    """Normalizes the image path for output."""
    if os.name == "nt":
        # On Windows, the JSON.dump ends up outputting un-escaped backslash
        # breaking the ability to read colors.json. Windows supports forward
        # slash, so we can use that for now
        return img.replace("\\", "/")
    return img


def colors_to_dict(colors, img):
    """Convert list of colors to pywal format."""
    return {
        "checksum": util.get_img_checksum(img),
        "wallpaper": normalize_img_path(img),
        "alpha": util.Color.alpha_num,
        "special": {
            "background": colors[0],
            "foreground": colors[15],
            "cursor": colors[15],
        },
        "colors": {
            "color0": colors[0],
            "color1": colors[1],
            "color2": colors[2],
            "color3": colors[3],
            "color4": colors[4],
            "color5": colors[5],
            "color6": colors[6],
            "color7": colors[7],
            "color8": colors[8],
            "color9": colors[9],
            "color10": colors[10],
            "color11": colors[11],
            "color12": colors[12],
            "color13": colors[13],
            "color14": colors[14],
            "color15": colors[15],
        },
    }


def shade_16(colors, light, cols16):
    """Generic 16 color shading
    this function will apply the 16 color shading
    to any color dict it is passed

    colors: dict
    light:  boolean - werether the colorscheme is light
    cols16: str [lighten|darken] - method to generate the shades"""

    # detect dict type
    if "color0" in colors:
        k_v = [
                "color0", "color1", "color2", "color3",
                "color4", "color5", "color6", "color7",
                "color8", "color9", "color10", "color11",
                "color12", "color13", "color14", "color15",
              ]
    else:
        k_v = [
                0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
              ]

    if cols16:
        if light:
            colors[k_v[7]] = util.darken_color(colors[k_v[0]], 0.50)
            colors[k_v[8]] = util.darken_color(colors[k_v[0]], 0.25)
            if cols16 == "lighten":
                colors[k_v[9]] = util.lighten_color(colors[k_v[1]], 0.25)
                colors[k_v[10]] = util.lighten_color(colors[k_v[2]], 0.25)
                colors[k_v[11]] = util.lighten_color(colors[k_v[3]], 0.25)
                colors[k_v[12]] = util.lighten_color(colors[k_v[4]], 0.25)
                colors[k_v[13]] = util.lighten_color(colors[k_v[5]], 0.25)
                colors[k_v[14]] = util.lighten_color(colors[k_v[6]], 0.25)
                colors[k_v[15]] = util.darken_color(colors[k_v[0]], 0.75)
            else:
                colors[k_v[1]] = util.darken_color(colors[k_v[1]], 0.25)
                colors[k_v[2]] = util.darken_color(colors[k_v[2]], 0.25)
                colors[k_v[3]] = util.darken_color(colors[k_v[3]], 0.25)
                colors[k_v[4]] = util.darken_color(colors[k_v[4]], 0.25)
                colors[k_v[5]] = util.darken_color(colors[k_v[5]], 0.25)
                colors[k_v[6]] = util.darken_color(colors[k_v[6]], 0.25)
                colors[k_v[15]] = util.darken_color(colors[k_v[0]], 0.75)
        else:
            colors[k_v[7]] = util.lighten_color(colors[k_v[0]], 0.55)
            colors[k_v[7]] = util.saturate_color(colors[k_v[7]], 0.05)
            colors[k_v[8]] = util.lighten_color(colors[k_v[0]], 0.35)
            colors[k_v[8]] = util.saturate_color(colors[k_v[8]], 0.10)
            colors[k_v[15]] = util.lighten_color(colors[k_v[0]], 0.75)
            if cols16 == "lighten":
                colors[k_v[9]] = util.lighten_color(colors[k_v[1]], 0.25)
                colors[k_v[10]] = util.lighten_color(colors[k_v[2]], 0.25)
                colors[k_v[11]] = util.lighten_color(colors[k_v[3]], 0.25)
                colors[k_v[12]] = util.lighten_color(colors[k_v[4]], 0.25)
                colors[k_v[13]] = util.lighten_color(colors[k_v[5]], 0.25)
                colors[k_v[14]] = util.lighten_color(colors[k_v[6]], 0.25)
                for i in range(9, 15):
                    colors[k_v[i]] = util.saturate_color(colors[k_v[i]], 0.40)
            else:
                colors[k_v[1]] = util.darken_color(colors[k_v[1]], 0.25)
                colors[k_v[2]] = util.darken_color(colors[k_v[2]], 0.25)
                colors[k_v[3]] = util.darken_color(colors[k_v[3]], 0.25)
                colors[k_v[4]] = util.darken_color(colors[k_v[4]], 0.25)
                colors[k_v[5]] = util.darken_color(colors[k_v[5]], 0.25)
                colors[k_v[6]] = util.darken_color(colors[k_v[6]], 0.25)


def generic_adjust(colors, light, **kwargs):
    """Generic color adjustment for themers.
    :keyword-args:
    -    c16 - [ "lighten" | "darken" ]
    """
    if "c16" in kwargs:
        cols16 = kwargs["c16"]
    else:
        cols16 = False

    if light:
        for color in colors:
            color = util.saturate_color(color, 0.60)
            color = util.darken_color(color, 0.5)

        colors[0] = util.lighten_color(colors[0], 0.95)
        if cols16:
            shade_16(colors, light, cols16)
        else:
            colors[7] = util.darken_color(colors[0], 0.75)
            colors[8] = util.darken_color(colors[0], 0.25)
            colors[15] = colors[7]

    else:
        if colors[0][1] != "0":  # the color may already be dark enough
            colors[0] = util.darken_color(colors[0], 0.40)  # just a bit darker

        saturate_more = False
        if colors[0][1] == "0":  # the color may not be saturated enough
            saturate_more = True
        if colors[0][3] == "0":  # the color may not be saturated enough
            saturate_more = True
        if colors[0][5] == "0":  # the color may not be saturated enough
            saturate_more = True

        if saturate_more:
            colors[0] = util.lighten_color(colors[0], 0.03)
            colors[0] = util.saturate_color(colors[0], 0.40)

        if cols16:
            shade_16(colors, light, cols16)
        else:
            colors[7] = util.lighten_color(colors[0], 0.75)
            colors[8] = util.lighten_color(colors[0], 0.35)
            colors[8] = util.saturate_color(colors[8], 0.10)
            colors[15] = colors[7]

    return colors


def saturate_colors(colors, amount):
    """Saturate all colors."""
    if amount and float(amount) <= 1.0:
        for i, _ in enumerate(colors):
            if i not in [7, 15]:
                colors[i] = util.saturate_color(colors[i], float(amount))

    return colors


def ensure_contrast(colors, contrast, light, image):
    """Ensure user-specified W3 contrast of colors
    depending on dark or light theme."""
    # If no contrast checking was specified, do nothing
    if not contrast or contrast == "":
        return colors

    # Contrast must be within a predefined range
    if float(contrast) < 1 or float(contrast) > 21:
        logging.error("Specified contrast ratio is too extreme")
        return colors

    # Get the image background color
    background_color = util.Color(util.image_average_color(image))
    background_luminance = background_color.w3_luminance

    # Calculate the required W3 luminance for the desired contrast ratio
    # This will modify all of the colors to be brighter or darker than the
    # background image depending on whether the user has specified for a
    # dark or light theme
    try:
        if light:
            luminance_desired = (background_luminance + 0.05) / float(
                contrast
            ) - 0.05
        else:
            luminance_desired = (background_luminance + 0.05) * float(
                contrast
            ) - 0.05
    except ValueError:
        logging.error("ensure_contrast(): Contrast valued could not be parsed")
        return colors

    if luminance_desired >= 0.99:
        print("Can't contrast this palette without changing colors to white")
        return colors
    if luminance_desired <= 0.01:
        print("Can't contrast this palette without changing colors to black")
        return colors

    # Determine which colors should be modified / checked
    # ! For the time being this is just going to modify all the colors except
    # 0 and 15
    colors_to_contrast = range(1, 15)

    # Modify colors
    for index in colors_to_contrast:
        color = util.Color(colors[index])

        # If the color already has sufficient contrast, do nothing
        if light and color.w3_luminance <= luminance_desired:
            continue
        elif color.w3_luminance >= luminance_desired:
            continue

        h, s, v = colorsys.rgb_to_hsv(
            float(color.red), float(color.green), float(color.blue)
        )

        # Determine how to modify the color based on its HSV characteristics

        # If the color is to be lighter than background, and the HSV color
        # with value 1 has sufficient luminance, adjust by increasing value
        if (
            not light
            and util.Color(
                util.rgb_to_hex(
                    [
                        int(channel * 255)
                        for channel in colorsys.hsv_to_rgb(h, s, 1)
                    ]
                )
            ).w3_luminance
            >= luminance_desired
        ):
            colors[index] = binary_luminance_adjust(
                luminance_desired, h, s, s, v, 1
            )
        # If the color is to be lighter than background and increasing value
        # to 1 doesn't produce the desired luminance, additionally decrease
        # saturation
        elif not light:
            colors[index] = binary_luminance_adjust(
                luminance_desired, h, 0, s, 1, 1
            )
        # If the color is to be darker than background, produce desired
        # luminance by decreasing value, and raising saturation
        else:
            colors[index] = binary_luminance_adjust(
                luminance_desired, h, s, 1, 0, v
            )

    return colors


def binary_luminance_adjust(
    luminance_desired, hue, s_min, s_max, v_min, v_max, iterations=10
):
    """Use a binary method to adjust a color's value and/or
    saturation to produce the desired luminance"""
    for i in range(iterations):
        # Obtain a new color by averaging saturation and value
        s = (s_min + s_max) / 2
        v = (v_min + v_max) / 2

        # Compare the luminance of this color to the target luminance
        # If the color is too light, clamp the minimum saturation
        # and maximum value
        if (
            util.Color(
                util.rgb_to_hex(
                    [
                        int(channel * 255)
                        for channel in colorsys.hsv_to_rgb(hue, s, v)
                    ]
                )
            ).w3_luminance
            >= luminance_desired
        ):
            s_min = s
            v_max = v
        # If the color is too dark, clamp the maximum saturation
        # and minimum value
        else:
            s_max = s
            v_min = v

    return util.rgb_to_hex(
        [int(channel * 255) for channel in colorsys.hsv_to_rgb(hue, s, v)]
    )


def cache_fname(img, backend, light, cache_dir, sat="", **kwargs):
    """Create the cache file name.
    :keyword-args:
    -    c16: use 16 colors through specified method - [ "lighten" | "darken" ]
    -    cst: palette contrast ratio - float
    """
    color_type = "light" if light else "dark"
    if "c16" in kwargs:
        cols16 = kwargs["c16"]
    else:
        cols16 = False
    if "cst" in kwargs:
        contrast = kwargs["cst"]
    else:
        contrast = False
    color_num = "16" if cols16 else "9"
    file_name = re.sub("[/|\\|.]", "_", img)
    file_size = os.path.getsize(img)

    if cols16 and contrast:
        file_parts = [
            file_name,
            color_num,
            cols16,
            color_type,
            backend,
            sat,
            contrast,
            file_size,
            __cache_version__,
        ]
        return [
            cache_dir,
            "schemes",
            "%s_%s_%s_%s_%s_%s_%s_%s_%s.json" % (*file_parts,),
        ]
    if cols16 and (not contrast):
        file_parts = [
            file_name,
            color_num,
            cols16,
            color_type,
            backend,
            sat,
            file_size,
            __cache_version__,
        ]
        return [
            cache_dir,
            "schemes",
            "%s_%s_%s_%s_%s_%s_%s_%s.json" % (*file_parts,),
        ]
    if (not cols16) and contrast:
        file_parts = [
            file_name,
            color_type,
            backend,
            sat,
            contrast,
            file_size,
            __cache_version__,
        ]
        return [
            cache_dir,
            "schemes",
            "%s_%s_%s_%s_%s_%s_%s.json" % (*file_parts,),
        ]
    else:
        file_parts = [
            file_name,
            color_type,
            backend,
            sat,
            file_size,
            __cache_version__,
        ]
        return [
            cache_dir,
            "schemes",
            "%s_%s_%s_%s_%s_%s.json" % (*file_parts,),
        ]


def get_backend(backend):
    """Figure out which backend to use."""
    if backend == "random":
        backends = list_backends()
        random.shuffle(backends)
        return backends[0]

    return backend


def palette():
    """Generate a palette from the colors."""
    for i in range(0, 16):
        if i % 8 == 0:
            print()

        if i > 7:
            i = "8;5;%s" % i

        print("\033[4%sm%s\033[0m" % (i, " " * (80 // 20)), end="")

    print("\n")


def get(
    img,
    light=False,
    backend="wal",
    cache_dir=CACHE_DIR,
    sat="",
    **kwargs,
):
    """Generate a palette.
    :keyword-args:
    -    c16: use 16 colors through specified method - [ "lighten" | "darken" ]
    -    cst: apply contrast ratio to palette        - float
    """
    if "c16" in kwargs:
        cols16 = kwargs["c16"]
    else:
        cols16 = False
    if "cst" in kwargs:
        contrast = kwargs["cst"]
    else:
        contrast = ""

    # home_dylan_img_jpg_backend_1.2.2.json
    if not contrast or contrast == "":
        cache_name = cache_fname(
            img, backend, light, cache_dir, sat,
            c16=cols16
        )
    else:
        cache_name = cache_fname(
            img, backend, light, cache_dir, sat,
            c16=cols16, cst=float(contrast)
        )

    cache_file = os.path.join(*cache_name)

    # Check the wallpaper's checksum against the cache'
    if os.path.isfile(cache_file) and theme.parse(cache_file)[
        "checksum"
    ] == util.get_img_checksum(img):
        colors = theme.file(cache_file)
        logging.info("Found cached colorscheme.")

    else:
        logging.info("Generating a colorscheme.")
        backend = get_backend(backend)

        # Dynamically import the backend we want to use.
        # This keeps the dependencies "optional".
        try:
            __import__("pywal.backends.%s" % backend)
        except ImportError:
            __import__("pywal.backends.wal")
            backend = "wal"

        logging.info("Using %s backend.", backend)
        backend = sys.modules["pywal.backends.%s" % backend]
        colors = getattr(backend, "get")(img, light, c16=cols16)

        # Post-processing steps from command-line arguments
        colors = saturate_colors(colors, sat)
        colors = ensure_contrast(colors, contrast, light, img)

        colors = colors_to_dict(colors, img)

        util.save_file_json(colors, cache_file)
        logging.info("Generation complete.")

    return colors


def file(input_file):
    """Deprecated: symbolic link to --> theme.file"""
    return theme.file(input_file)
